"""Contains scientific validation methods on input signals."""
import efel


def check_ais_initiation(soma_trace: dict, ais_trace: dict) -> bool:
    """Checks the initiation of action potential in AIS with respect to soma."""
    f_values = efel.get_feature_values(
        traces=[soma_trace, ais_trace], feature_names=["AP_begin_time"],
        parallel_map=None, return_list=True, raise_warnings=True
    )
    soma_ap_begin_time = f_values[0]["AP_begin_time"]
    ais_ap_begin_time = f_values[1]["AP_begin_time"]

    if soma_ap_begin_time is None or ais_ap_begin_time is None:
        raise ValueError("AP_begin_time feature is not computed")

    if len(soma_ap_begin_time) != len(ais_ap_begin_time):
        raise ValueError("Number of APs in soma and AIS do not match")

    # Checking if no spike in the soma starts earlier than in the AIS
    for soma_time, ais_time in zip(soma_ap_begin_time, ais_ap_begin_time):
        if soma_time < ais_time:
            raise ValueError(
                "There is a spike that initiates in the soma before the axon"
            )

    return True



"""Features that are depending on the inter-spike intervals."""
from __future__ import annotations
import logging
from typing_extensions import deprecated
import warnings
import numpy as np
from efel.pyfeatures.cppfeature_access import (
    _get_cpp_data, get_cpp_feature
)


logger = logging.getLogger(__name__)


@deprecated("Use all_ISI_values instead")
def ISIs() -> np.ndarray | None:
    """Get all ISIs, inter-spike intervals."""
    return get_cpp_feature("all_ISI_values")


@deprecated("Use ISIs instead.")
def ISI_values() -> np.ndarray | None:
    """Get all ISIs, inter-spike intervals."""
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]
    return isi_values


def __ISI_CV(isi_values) -> float | None:
    if len(isi_values) < 2:
        return None

    # Calculate mean
    isi_mean = np.mean(isi_values)
    # Calculate coefficient of variation
    cv = np.std(isi_values, ddof=1) / isi_mean  # ddof 1 to replicate C++ impl
    return cv


def ISI_CV() -> np.ndarray | None:
    """Coefficient of variation of ISIs.

    If the ignore_first_ISI flag is set, the first ISI will be ignored.
    """
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    result = __ISI_CV(isi_values)
    if result is None:
        return None
    return np.array([result])


def single_burst_ratio() -> np.ndarray | None:
    """Calculates the single burst ratio.

    The ratio is the length of the first ISI over the average of the rest.
    If the ignore_first_ISI flag is set, the first ISI will be ignored.
    """
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    if len(isi_values) < 2:
        return None

    single_burst_ratio_value = isi_values[0] / np.mean(isi_values)
    return np.array([single_burst_ratio_value])


def irregularity_index() -> np.ndarray | None:
    """Calculate the irregularity index of ISI values.

    If the ignore_first_ISI flag is set, the first ISI will be ignored.
    """
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    # Calculate the absolute differences between consecutive ISI values
    isi_differences = np.abs(np.diff(isi_values))
    result = np.mean(isi_differences)

    return np.array([result])


def _isi_log_slope_core(
    isi_values, skip=False, spike_skipf=0.0, max_spike_skip=0, semilog=False
) -> np.ndarray | None:
    if isi_values is None or len(isi_values) == 0:
        return None

    if skip:
        isisToRemove = min(
            max_spike_skip, int((len(isi_values) + 1) * spike_skipf + 0.5)
        )
        isi_values = isi_values[isisToRemove:]

    log_isi_values = np.log(isi_values)
    x = np.arange(1, len(log_isi_values) + 1)
    if not semilog:
        x = np.log(x)

    try:
        slope, _ = np.polyfit(x, log_isi_values, 1)
    except (np.linalg.LinAlgError, SystemError) as e:
        warnings.warn(f"Error in polyfit: {e}")
        return None

    return np.array([slope])


def ISI_log_slope() -> np.ndarray | None:
    """The slope of a linear fit to a loglog plot of the ISI values.

    If the ignore_first_ISI flag is set, the first ISI will be ignored.
    """
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    return _isi_log_slope_core(isi_values, False, 0.0, 0, False)


def ISI_semilog_slope() -> np.ndarray | None:
    """The slope of a linear fit to a semilog plot of the ISI values.

    If the ignore_first_ISI flag is set, the first ISI will be ignored.
    """
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None

    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    return _isi_log_slope_core(isi_values, False, 0.0, 0, True)


def ISI_log_slope_skip() -> np.ndarray | None:
    """The slope of a linear fit to a loglog plot of the ISI values,
    but not taking into account the first ISI values.

    Uses the spike_skipf and max_spike_skip settings to determine how many
    ISIs to skip.
    ."""
    isi_values = get_cpp_feature("all_ISI_values")
    if isi_values is None:
        return None
    # Check "ignore_first_ISI" flag
    ignore_first_ISI = _get_cpp_data("ignore_first_ISI")
    if ignore_first_ISI:
        isi_values = isi_values[1:]

    spike_skipf = _get_cpp_data("spike_skipf")
    if spike_skipf < 0 or spike_skipf >= 1:
        raise ValueError("spike_skipf should lie between [0, 1).")
    max_spike_skip = _get_cpp_data("max_spike_skip")
    return _isi_log_slope_core(isi_values, True, spike_skipf, max_spike_skip, False)


def burst_ISI_indices() -> np.ndarray | None:
    """Calculate burst ISI indices based on burst factor and ISI values."""
    # Fetching necessary data
    isi_values = ISI_values()
    if isi_values is None:
        return None

    burst_factor = _get_cpp_data("burst_factor")

    if len(isi_values) < 4:
        logger.warning("4 or more spikes are needed for burst calculation.")
        return None

    burst_indices = []
    count = -1

    for i in range(1, len(isi_values) - 1):
        isi_p_copy = isi_values[count + 1: i]
        n = len(isi_p_copy)

        if n == 0:
            continue

        # Compute median
        d_median = np.median(isi_p_copy)

        # Check burst condition
        if isi_values[i] > (burst_factor * d_median) and isi_values[i + 1] < (
            isi_values[i] / burst_factor
        ):
            burst_indices.append(i + 1)
            count = i - 1

    if burst_indices == []:
        return None
    return np.array(burst_indices)


def burst_mean_freq() -> np.ndarray | None:
    """Calculate the mean frequency of bursts."""
    # Fetching required features
    peak_time = get_cpp_feature("peak_time")
    burst_isi_idx = burst_ISI_indices()
    if burst_isi_idx is None or peak_time is None:
        return None

    burst_mean_freq = []
    burst_index_tmp = burst_isi_idx
    burst_index = np.insert(
        burst_index_tmp, burst_index_tmp.size, len(peak_time) - 1
    )
    burst_index = burst_index.astype(int)

    # 1st burst
    span = peak_time[burst_index[0]] - peak_time[0]
    # + 1 because 1st ISI is ignored
    N_peaks = burst_index[0] + 1
    burst_mean_freq.append(N_peaks * 1000 / span)

    for i, burst_idx in enumerate(burst_index[:-1]):
        if burst_index[i + 1] - burst_idx != 1:
            span = peak_time[burst_index[i + 1]] - peak_time[burst_idx + 1]
            N_peaks = burst_index[i + 1] - burst_idx
            burst_mean_freq.append(N_peaks * 1000 / span)

    return np.array(burst_mean_freq)


def interburst_voltage() -> np.ndarray | None:
    """The voltage average in between two bursts."""
    peak_idx = get_cpp_feature("peak_indices")
    v = get_cpp_feature("voltage")
    t = get_cpp_feature("time")
    burst_isi_idx = burst_ISI_indices()
    if peak_idx is None or v is None or t is None or burst_isi_idx is None:
        return None

    interburst_voltage = []
    for idx in burst_isi_idx:
        ts_idx = peak_idx[idx]
        t_start = t[ts_idx] + 5
        start_idx = np.argwhere(t < t_start)[-1][0]

        te_idx = peak_idx[idx + 1]
        t_end = t[te_idx] - 5
        end_idx = np.argwhere(t > t_end)[0][0]

        interburst_voltage.append(np.mean(v[start_idx:end_idx + 1]))

    return np.array(interburst_voltage)


def initburst_sahp() -> np.ndarray | None:
    """SlowAHP voltage after initial burst."""
    # Required cpp features
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")
    if voltage is None or time is None:
        return None
    time = time[:len(voltage)]
    peak_times = get_cpp_feature("peak_time")
    if peak_times is None:
        return None

    # Required python features
    all_isis = get_cpp_feature("all_ISI_values")

    # Required trace data
    stim_end = _get_cpp_data("stim_end")

    # Required settings
    initburst_freq_thresh = _get_cpp_data("initburst_freq_threshold")
    initburst_sahp_start = _get_cpp_data("initburst_sahp_start")
    initburst_sahp_end = _get_cpp_data("initburst_sahp_end")

    last_isi = None

    # Loop over ISIs until frequency higher than initburst_freq_threshold
    if all_isis is None:
        return None
    for isi_counter, isi in enumerate(all_isis):
        # Convert to Hz
        freq = 1000.0 / isi
        if freq < initburst_freq_thresh:
            # Threshold reached
            break
        else:
            # Add isi to initburst
            last_isi = isi_counter

    if last_isi is None:
        # No initburst found
        return None
    else:
        # Get index of second peak of last ISI
        last_peak = last_isi + 1

    # Get time of last peak
    last_peak_time = peak_times[last_peak]

    # Determine start of sahp interval
    sahp_interval_start = min(
        last_peak_time +
        initburst_sahp_start,
        stim_end)

    # Get next peak, we wont search beyond that
    next_peak = last_peak + 1

    # Determine end of sahp interval
    # Add initburst_slow_ahp_max to last peak time
    # If next peak or stim_end is earlier, use these
    # If no next peak, use stim end
    if next_peak < len(peak_times):
        next_peak_time = peak_times[next_peak]

        sahp_interval_end = min(
            last_peak_time + initburst_sahp_end, next_peak_time, stim_end)
    else:
        sahp_interval_end = min(
            last_peak_time + initburst_sahp_end, stim_end)

    if sahp_interval_end <= sahp_interval_start:
        return None
    else:
        sahp_interval = voltage[np.where(
            (time <= sahp_interval_end) &
            (time >= sahp_interval_start))]

        if len(sahp_interval) > 0:
            min_volt_index = np.argmin(sahp_interval)
        else:
            return None

        slow_ahp = sahp_interval[min_volt_index]

        return np.array([slow_ahp])


def strict_burst_number() -> np.ndarray:
    """Calculate the strict burst number.

    This implementation does not assume that every spike belongs to a burst.
    The first spike is ignored by default. This can be changed by setting
    ignore_first_ISI to 0.

    The burst detection can be fine-tuned by changing the setting
    strict_burst_factor. Default value is 2.0."""
    burst_mean_freq = get_cpp_feature("strict_burst_mean_freq")
    if burst_mean_freq is None:
        return np.array([0])
    return np.array([burst_mean_freq.size])


def inv_ISI_values() -> np.ndarray | None:
    """Calculate the inverse of ISI values."""
    all_isi_values_vec = get_cpp_feature("all_ISI_values")
    return None if all_isi_values_vec is None else 1000.0 / all_isi_values_vec


def inv_first_ISI() -> np.ndarray | None:
    """Calculate the inverse of the first ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) == 0:
        return None
    return np.array([inv_isi_values[0]])


def inv_second_ISI() -> np.ndarray | None:
    """Calculate the inverse of the second ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) < 2:
        return None
    return np.array([inv_isi_values[1]])


def inv_third_ISI() -> np.ndarray | None:
    """Calculate the inverse of the third ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) < 3:
        return None
    return np.array([inv_isi_values[2]])


def inv_fourth_ISI() -> np.ndarray | None:
    """Calculate the inverse of the fourth ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) < 4:
        return None
    return np.array([inv_isi_values[3]])


def inv_fifth_ISI() -> np.ndarray | None:
    """Calculate the inverse of the fifth ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) < 5:
        return None
    return np.array([inv_isi_values[4]])


def inv_last_ISI() -> np.ndarray | None:
    """Calculate the inverse of the last ISI."""
    inv_isi_values = inv_ISI_values()
    if inv_isi_values is None or len(inv_isi_values) == 0:
        return None
    return np.array([inv_isi_values[-1]])



"""Contains the features that are computed using multiple traces."""


import numpy as np

import efel


def bpap_attenuation(soma_trace: dict, dendrite_trace: dict) -> float:
    """Computes the attenuation of backpropagating action potential.

    Backpropagating action potential is the action potential that is initiated
    in the soma and propagates to the dendrite. The attenuation is the ratio
    of the amplitude of the action potential in the soma and the dendrite.
    The attenuation is computed by first subtracting the resting potential
    from the voltage traces.
    """
    f_values = efel.get_feature_values(
        [soma_trace, dendrite_trace], ["voltage_base"],
        parallel_map=None, return_list=True, raise_warnings=True)
    vb_soma = f_values[0]["voltage_base"][0]
    vb_dend = f_values[1]["voltage_base"][0]
    v_soma = soma_trace["V"]
    v_dend = dendrite_trace["V"]
    res = (np.max(v_soma) - vb_soma) / (np.max(v_dend) - vb_dend)
    return res



from __future__ import annotations
"""IO handler for eFEL"""

"""
Copyright (c) 2015, EPFL/Blue Brain Project

 This file is part of eFEL <https://github.com/BlueBrain/eFEL>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
import csv
import json
from pathlib import Path
import neo
import numpy as np


def load_ascii_input(
    file_path: Path | str, delimiter: str = " "
) -> tuple[np.ndarray, np.ndarray]:
    """Loads electrophysiology data from an ASCII file.

    Returns: A tuple containing two numpy arrays, one for time and one for voltage.
    """
    file_path = Path(file_path)
    data = np.loadtxt(file_path, delimiter=delimiter)
    time, voltage = data[:, 0], data[:, 1]
    return time, voltage


def extract_stim_times_from_neo_data(blocks, stim_start, stim_end) -> tuple:
    """
    Seeks for the stim_start and stim_end parameters inside the Neo data.

    Args:
        blocks (Neo object blocks): Description of what blocks represents.
        stim_start (numerical value or None): Start time of the stimulation in
            milliseconds. If not available, None should be used.
        stim_end (numerical value or None): End time of the stimulation in
            milliseconds. If not available, None should be used.

    Returns:
        tuple: A tuple containing:
            - stim_start (numerical value or None): Start time of the stimulation
              in milliseconds.
            - stim_end (numerical value or None): End time of the stimulation in
              milliseconds.

    Notes:
        - Epoch.name should be one of "stim", "stimulus", "stimulation",
          "current_injection".
        - First Event.name should be "stim_start", "stimulus_start",
          "stimulation_start", "current_injection_start".
        - Second Event.name should be one of "stim_end", "stimulus_end",
          "stimulation_end", "current_injection_end".

    """
    # this part code aims to find informations about stimulations, if
    # stimulation time are None.
    if stim_start is None and stim_end is None:
        for bl in blocks:
            for seg in bl.segments:
                event_start_rescaled = False
                event_end_rescaled = False
                epoch_rescaled = False
                for epoch in seg.epochs:
                    if epoch.name in (
                            "stim",
                            "stimulus",
                            "stimulation",
                            "current_injection"):
                        if stim_start is None:
                            epoch = epoch.rescale('ms').magnitude
                            stim_start = epoch[0]
                            stim_end = epoch[-1]
                            epoch_rescaled = True
                        else:
                            raise ValueError(
                                'It seems that there are two epochs related '
                                'to stimulation, the program does not know '
                                'which one to chose')

                for event in seg.events:
                    # Test event stim_start
                    if event.name in (
                            "stim_start",
                            "stimulus_start",
                            "stimulation_start",
                            "current_injection_start"):
                        # tests if not already found once
                        if event_start_rescaled:
                            raise ValueError(
                                'It seems that stimulation start time is '
                                'defined more than one event.'
                                ' The program does not know which one '
                                'to choose')
                        if epoch_rescaled:
                            raise ValueError(
                                'It seems that stim_start is defined in both '
                                'epoch and an event.' +
                                ' The program does not know which one '
                                'to choose')

                        if stim_start is None:
                            event = event.rescale('ms').magnitude

                            try:
                                stim_start = event[0]
                            except BaseException:
                                stim_start = event

                            event_start_rescaled = True

                    # Test event stim_end
                    elif event.name in (
                        "stim_end",
                        "stimulus_end",
                        "stimulation_end",
                        "current_injection_end",
                    ):
                        # tests if not already found once
                        if event_end_rescaled:
                            raise ValueError(
                                'It seems that stimulation end time is '
                                'defined more than one event.'
                                ' The program does not know which one '
                                'to choose')
                        if epoch_rescaled:
                            raise ValueError(
                                'It seems that stim_end is defined in '
                                'both epoch and an event.'
                                ' The program does not know which one '
                                'to choose')

                        if stim_end is None:
                            event = event.rescale('ms').magnitude

                            try:
                                stim_end = event[-1]
                            except BaseException:
                                stim_end = event

                            event_end_rescaled = True

    return stim_start, stim_end


def load_neo_file(file_name: str, stim_start=None, stim_end=None, **kwargs) -> list:
    """
    Loads a data file using neo and converts it for eFEL readability.

    Args:
        file_name (string): Path to the Dependency file location.
        stim_start (numerical value, optional): Start time in ms. Optional if an Epoch
            or two Events are in the file.
        stim_end (numerical value, optional): End time in ms. Optional if an Epoch
            or two Events are in the file.
        **kwargs: Additional arguments for the read() method of Neo IO class.

    Returns:
        list of Segments: Segments containing traces, formatted as
            [Segments_1, Segments_2, ..., Segments_n], where each Segments_i is
            [Traces_1, Traces_2, ..., Traces_n].

    Notes:
        - Epoch.name should be "stim", "stimulus", "stimulation", "current_injection".
        - First Event.name: "stim_start", "stimulus_start", "stimulation_start",
          "current_injection_start".
        - Second Event.name: "stim_end", "stimulus_end", "stimulation_end",
          "current_injection_end".
    """
    reader = neo.io.get_io(file_name)
    blocks = reader.read(**kwargs)

    stim_start, stim_end = extract_stim_times_from_neo_data(
        blocks, stim_start, stim_end)
    if stim_start is None or stim_end is None:
        raise ValueError(
            'No stim_start or stim_end found in epochs or events. '
            'Please specify "stim_start" and "stim_end" arguments.')

    # Convert data for eFEL
    efel_blocks = []
    for bl in blocks:
        efel_segments = []
        for seg in bl.segments:
            traces = []
            for sig in seg.analogsignals:
                if 'V' in sig.units.dimensionality.string:
                    trace = {
                        'T': (sig.times - sig.times[0]).rescale('ms').magnitude,
                        'V': sig.rescale('mV').magnitude,
                        'stim_start': [stim_start],
                        'stim_end': [stim_end]
                    }
                    traces.append(trace)
            efel_segments.append(traces)
        efel_blocks.append(efel_segments)

    return efel_blocks


def save_feature_to_json(feature_values, filename):
    """Save feature values as a JSON file."""
    class NumpyEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, np.integer):
                return int(o)
            elif isinstance(o, np.floating):
                return float(o)
            elif isinstance(o, np.ndarray):
                return o.tolist()
            else:
                return super().default(o)

    with open(filename, 'w') as f:
        json.dump(feature_values, f, cls=NumpyEncoder)


def save_feature_to_csv(feature_values, filename):
    """Save feature values as a CSV file."""
    with open(filename, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(feature_values.keys())
        max_list_length = max(len(value) if isinstance(value, (list, np.ndarray))
                              else 1 for value in feature_values.values())
        for i in range(max_list_length):
            row_data = [value[i] if isinstance(value, (list, np.ndarray))
                        and i < len(value) else ''
                        for value in feature_values.values()]
            writer.writerow(row_data)



"""Module to get units of efeatures."""
import json
import pkgutil


_units_raw = pkgutil.get_data(__name__, "units.json")

if _units_raw is None:
    raise ValueError("Failed to load units.json")

_units = json.loads(_units_raw)


def get_unit(feature_name: str) -> str:
    """Get the unit of a feature."""
    return _units[feature_name]



"""Extracellular features functions"""

"""
Copyright (c) 2024, EPFL/Blue Brain Project

 This file is part of eFEL <https://github.com/BlueBrain/eFEL>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy as np
from scipy.stats import linregress
from scipy.signal import resample_poly

all_1D_features = [
    "peak_to_valley",
    "halfwidth",
    "peak_trough_ratio",
    "repolarization_slope",
    "recovery_slope",
    "neg_peak_relative",
    "pos_peak_relative",
    "neg_peak_diff",
    "pos_peak_diff",
    "neg_image",
    "pos_image",
]


def _get_slope(x, y):
    """
    Return the slope of x and y data, using scipy.signal.linregress
    """

    slope = linregress(x, y)
    return slope


def _get_trough_and_peak_idx(waveform, after_max_trough=False):
    """
    Return the indices of the detected troughs (minimum of waveform)
    and peaks (maximum of waveform, after trough) of the input waveforms.

    Assumes negative troughs and positive peaks

    Returns 0 if not detected
    """
    if after_max_trough:
        max_trough_idx = np.unravel_index(
            np.argmin(waveform),
            waveform.shape)[1]
        trough_idx = (
            np.argmin(waveform[:, max_trough_idx:], axis=1) + max_trough_idx
        )
        peak_idx = (
            np.argmax(waveform[:, max_trough_idx:], axis=1) + max_trough_idx
        )
    else:
        trough_idx = np.argmin(waveform, axis=1)
        peak_idx = np.argmax(waveform, axis=1)

    return trough_idx, peak_idx


def calculate_features(
    waveforms,
    sampling_frequency,
    upsample=None,
    feature_names=None,
    recovery_slope_window=0.7
):
    """Calculate features for all waveforms

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sampling_frequency  : float
            rate at which the waveforms are sampled (Hz)
        feature_names : list or None (if None, compute all)
            features to compute
        recovery_slope_window : float
            window length in ms after peak wherein recovery slope is computed

    Returns:
        dict (num_waveforms x num_metrics): Dictionary with computed metrics.
            Keys are the metric names, values are the computed features
    """
    metrics = dict()

    if feature_names is None:
        feature_names = all_1D_features
    else:
        for name in feature_names:
            assert name in all_1D_features, f"{name} not in {all_1D_features}"

    if upsample is not None:
        assert upsample > 0
        waveforms = _upsample_wf(waveforms, int(upsample))
        sampling_frequency = upsample * sampling_frequency

    if "peak_to_valley" in feature_names:
        metrics["peak_to_valley"] = peak_to_valley(
            waveforms=waveforms, sampling_frequency=sampling_frequency
        )
    if "peak_trough_ratio" in feature_names:
        metrics["peak_trough_ratio"] = peak_trough_ratio(waveforms=waveforms)

    if "halfwidth" in feature_names:
        metrics["halfwidth"] = halfwidth(
            waveforms=waveforms, sampling_frequency=sampling_frequency
        )

    if "repolarization_slope" in feature_names:
        metrics["repolarization_slope"] = repolarization_slope(
            waveforms=waveforms,
            sampling_frequency=sampling_frequency,
        )

    if "recovery_slope" in feature_names:
        metrics["recovery_slope"] = recovery_slope(
            waveforms=waveforms,
            sampling_frequency=sampling_frequency,
            window=recovery_slope_window,
        )

    if "neg_peak_diff" in feature_names:
        metrics["neg_peak_diff"] = peak_time_diff(
            waveforms=waveforms, fs=sampling_frequency, sign="negative"
        )

    if "pos_peak_diff" in feature_names:
        metrics["pos_peak_diff"] = peak_time_diff(
            waveforms=waveforms, fs=sampling_frequency, sign="positive"
        )

    if "neg_peak_relative" in feature_names:
        metrics["neg_peak_relative"] = relative_amplitude(
            waveforms=waveforms, sign="negative"
        )

    if "pos_peak_relative" in feature_names:
        metrics["pos_peak_relative"] = relative_amplitude(
            waveforms=waveforms, sign="positive"
        )

    if "neg_image" in feature_names:
        metrics["neg_image"] = peak_image(waveforms=waveforms, sign="negative")

    if "pos_image" in feature_names:
        metrics["pos_image"] = peak_image(waveforms=waveforms, sign="positive")

    return metrics


def peak_to_valley(waveforms, sampling_frequency):
    """
    Time between trough and peak. If the peak precedes the trough,
    peak_to_valley is negative.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute feature for
        sampling_frequency  : float
            rate at which the waveforms are sampled (Hz)

    Returns:
        np.ndarray (num_waveforms): peak_to_valley in seconds
    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    ptv = (peak_idx - trough_idx) * (1 / sampling_frequency)
    ptv[ptv == 0] = np.nan
    return ptv


def peak_trough_ratio(waveforms):
    """
    Normalized ratio of peak height over trough depth

    Assumes baseline is 0

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute feature for

    Returns:
        np.ndarray (num_waveforms): Peak to trough ratio
    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    ptratio = np.empty(trough_idx.shape[0])
    ptratio[:] = np.nan
    for i in range(waveforms.shape[0]):
        if peak_idx[i] == 0 and trough_idx[i] == 0:
            continue
        ptratio[i] = np.abs(waveforms[i, peak_idx[i]] /
                            waveforms[i, trough_idx[i]])

    return ptratio


def halfwidth(
    waveforms,
    sampling_frequency,
    return_idx=False
):
    """
    Width of waveform at half of its amplitude.
    If the peak precedes the trough, halfwidth is negative.

    Computes the width of the waveform peak at half its height

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sampling_frequency  : float
            rate at which the waveforms are sampled (Hz)
        return_idx : bool
            if true, also returns index of threshold crossing before and
            index of threshold crossing after peak

    Returns:
        np.ndarray or (np.ndarray, np.ndarray, np.ndarray):
            Halfwidth of the waveforms or (Halfwidth of the waveforms,
            index_cross_pre_peak, index_cross_post_peak)
    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)
    hw = np.empty(waveforms.shape[0])
    hw[:] = np.nan
    cross_pre_pk = np.empty(waveforms.shape[0], dtype=int)
    cross_post_pk = np.empty(waveforms.shape[0], dtype=int)

    for i in range(waveforms.shape[0]):
        if peak_idx[i] >= trough_idx[i]:
            trough_val = waveforms[i, trough_idx[i]]
            threshold = (
                0.5 * trough_val
            )  # threshold is half of peak heigth (assuming baseline is 0)

            cpre_idx = np.where(waveforms[i, :trough_idx[i]] < threshold)[0]
            cpost_idx = np.where(waveforms[i, trough_idx[i]:] < threshold)[0]

            if len(cpre_idx) == 0 or len(cpost_idx) == 0:
                continue

            cross_pre_pk[i] = (
                cpre_idx[0] - 1
            )  # last occurence of waveform lower than thr, before peak
            cross_post_pk[i] = (
                cpost_idx[-1] + 1 + trough_idx[i]
            )  # first occurence of waveform lower than peak, after peak

            hw[i] = (cross_post_pk[i] - cross_pre_pk[i]) * (
                1 / sampling_frequency
            )  # + peak_idx[i]
        else:
            peak_val = waveforms[i, peak_idx[i]]
            threshold = (
                0.5 * peak_val
            )  # threshold is half of peak heigth (assuming baseline is 0)

            cpre_idx = np.where(waveforms[i, :peak_idx[i]] > threshold)[0]
            cpost_idx = np.where(waveforms[i, peak_idx[i]:] > threshold)[0]

            if len(cpre_idx) == 0 or len(cpost_idx) == 0:
                continue

            cross_pre_pk[i] = (
                cpre_idx[0] - 1
            )  # last occurence of waveform lower than thr, before peak
            cross_post_pk[i] = (
                cpost_idx[-1] + 1 + trough_idx[i]
            )  # first occurence of waveform lower than peak, after peak

            hw[i] = -(cross_post_pk[i] - cross_pre_pk[i]) * (
                1 / sampling_frequency
            )  # + peak_idx[i]

    if not return_idx:
        return hw

    return hw, cross_pre_pk, cross_post_pk


def repolarization_slope(waveforms,
                         sampling_frequency,
                         return_idx=False
                         ):
    """
    Return slope of repolarization period between trough and baseline

    After reaching its maxumum polarization, the neuron potential will
    recover. The repolarization slope is defined as the dV/dT of the action
    potential between trough and baseline.

    Optionally the function returns also the indices per waveform where the
    potential crosses baseline.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sampling_frequency  : float
            rate at which the waveforms are sampled (Hz)
        return_idx : bool
            if true, also returns index of threshold crossing before and
            index of threshold crossing after peak

    Returns:
        np.ndarray or (np.ndarray, np.ndarray): Repolarization slope of the
            waveforms or (Repolarization slope of the waveforms, return to base
            index)
    """
    trough_idx, peak_idx = _get_trough_and_peak_idx(waveforms)

    rslope = np.empty(waveforms.shape[0])
    rslope[:] = np.nan
    return_to_base_idx = np.empty(waveforms.shape[0], dtype=np.int_)
    return_to_base_idx[:] = 0

    time = np.arange(0, waveforms.shape[1]) * (1 / sampling_frequency)  # in s
    for i in range(waveforms.shape[0]):
        if trough_idx[i] == 0:
            continue

        rtrn_idx = np.where(waveforms[i, trough_idx[i]:] >= 0)[0]
        if len(rtrn_idx) == 0:
            continue

        return_to_base_idx[i] = (
            rtrn_idx[0] + trough_idx[i]
        )  # first time after  trough, where waveform is at baseline

        if return_to_base_idx[i] - trough_idx[i] < 3:
            continue
        slope = _get_slope(
            time[trough_idx[i]:return_to_base_idx[i]],
            waveforms[i, trough_idx[i]:return_to_base_idx[i]]
        )
        rslope[i] = slope[0]

    if not return_idx:
        return rslope

    return rslope, return_to_base_idx


def recovery_slope(waveforms, sampling_frequency, window):
    """
    Return the recovery slope of input waveforms. After repolarization,
    the neuron hyperpolarizes until it peaks. The recovery slope is the
    slope of the action potential after the peak, returning to the baseline
    in dV/dT. The slope is computed within a user-defined window after
    the peak.

    Takes a numpy array of waveforms and returns an array with
    recovery slopes per waveform.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sampling_frequency  : float
            rate at which the waveforms are sampled (Hz)
        window : float
            length after peak wherein to compute recovery slope (ms)

    Returns:
        np.ndarray: Recovery slope of the waveforms
    """
    _, peak_idx = _get_trough_and_peak_idx(waveforms)
    rslope = np.empty(waveforms.shape[0])
    rslope[:] = np.nan

    time = np.arange(0, waveforms.shape[1]) * (1 / sampling_frequency)  # in s

    for i in range(waveforms.shape[0]):
        if peak_idx[i] in [0, waveforms.shape[1]]:
            continue
        max_idx = int(peak_idx[i] + ((window / 1000) * sampling_frequency))
        max_idx = np.min([max_idx, waveforms.shape[1]])

        if len(time[peak_idx[i]:max_idx]) < 3:
            continue
        slope = _get_slope(
            time[peak_idx[i]:max_idx], waveforms[i, peak_idx[i]:max_idx]
        )
        rslope[i] = slope[0]

    return rslope


def peak_image(waveforms, sign="negative"):
    """
    Normalized amplitude at the time of peak minimum or maximum.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sign : str
            "positive" | "negative"

    Returns:
        np.ndarray: Peak images for the waveforms
    """
    assert len(waveforms) > 1

    if sign == "negative":
        funarg = np.argmin
        fun = np.min
    else:
        funarg = np.argmax
        fun = np.max

    peak_channel, peak_time = np.unravel_index(
        funarg(waveforms), waveforms.shape
    )
    relative_peaks = waveforms[:, peak_time] / fun(waveforms[peak_channel])

    return relative_peaks


def relative_amplitude(waveforms, sign="negative"):
    """
    Normalized amplitude with respect to channel with largest amplitude.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        sign : str
            "positive" | "negative"

    Returns:
        np.ndarray: Relative amplitudes for the waveforms
    """
    assert len(waveforms) > 1

    if sign == "negative":
        fun = np.min
    else:
        fun = np.max

    peak_amp = np.abs(fun(waveforms))
    relative_peaks = np.abs(fun(waveforms, 1)) / peak_amp

    return relative_peaks


def peak_time_diff(waveforms, fs, sign="negative"):
    """
    Peak time differences with respect to channel with largest amplitude.

    Args:
        waveforms  : numpy.ndarray (num_waveforms x num_samples)
            waveforms to compute features for
        fs : float
            Sampling rate in Hz
        sign : str
            "positive" | "negative"

    Returns:
        np.ndarray: Peak time differences for the waveforms
    """
    assert len(waveforms) > 1

    if sign == "negative":
        argfun = np.argmin
    else:
        argfun = np.argmax

    peak_chan = np.unravel_index(argfun(waveforms), waveforms.shape)[0]
    peak_time = argfun(waveforms[peak_chan])
    relative_peak_times = (argfun(waveforms, 1) - peak_time) / fs

    return relative_peak_times


def _upsample_wf(waveforms, upsample):

    ndim = len(waveforms.shape)
    waveforms_up = resample_poly(waveforms, up=upsample, down=1, axis=ndim - 1)

    return waveforms_up



"""eFEL Python API functions.

This module provides the user-facing Python API of eFEL.
The convenience functions defined here call the underlying 'cppcore' library
to hide the lower level API from the user.


Copyright (c) 2015, EPFL/Blue Brain Project

 This file is part of eFEL <https://github.com/BlueBrain/eFEL>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# pylint: disable=W0602,W0603,W0702, F0401, W0612, R0912
from __future__ import annotations

from pathlib import Path
from typing import Callable, Iterator, Literal, overload
from typing_extensions import deprecated
import numpy as np

import efel
import efel.cppcore as cppcore

import efel.pyfeatures as pyfeatures
from efel.pyfeatures.pyfeatures import get_cpp_feature

_settings = efel.Settings()


def set_setting(setting_name: str, new_value: int | float | str) -> None:
    """Set a certain setting to a new value.

    Args:
        setting_name: Name of the setting to change.
        new_value: New value for the setting.
    """
    _settings.set_setting(setting_name, new_value)


def get_settings() -> efel.Settings:
    """Returns the current settings of eFEL."""
    return _settings


def reset():
    """Resets the efel settings to their default values.
    see :py:func:`efel.Settings`
    """
    global _settings
    _settings = efel.Settings()
    _settings.reset_to_default()
    _initialise()


@deprecated("Use `set_setting('dependencyfile_path', location)` instead")
def set_dependency_file_location(location: str | Path) -> None:
    """Sets the location of the Dependency file.

    eFEL uses 'Dependency' files to let the user define versions of features to use.
    The installation directory of eFEL contains a default 'DependencyV5.txt' file.
    Unless users want to change this file, it is not necessary to call this function.
    Modifying the Dependency file can be useful in debugging.

    Args:
        location: Path to the location of a Dependency file.

    Raises:
        FileNotFoundError: If the path to the dependency file doesn't exist.
    """
    set_setting('dependencyfile_path', str(location))


def get_dependency_file_location() -> str:
    """Gets the location of the Dependency file.

    Returns:
        Path to the location of a Dependency file.
    """
    return _settings.dependencyfile_path


@deprecated("Use `set_setting('Threshold', new_threshold)` instead")
def set_threshold(new_threshold: float) -> None:
    """Set the spike detection threshold in the eFEL, default -20.0

    Args:
        new_threshold: The new spike detection threshold value (in the same units
                       as the traces, e.g. mV).
    """
    set_setting('Threshold', new_threshold)


@deprecated("Use `set_setting('DerivativeThreshold', "
            "new_derivative_threshold)` instead")
def set_derivative_threshold(new_derivative_threshold: float) -> None:
    """Set the threshold for the derivative for detecting the spike onset.

    Some features use a threshold on dV/dt to calculate the beginning of an
    action potential. This function allows you to set this threshold.

    Args:
        new_derivative_threshold: The new derivative threshold value (in the same units
                                  as the traces, e.g. mV/ms).
    """
    set_setting('DerivativeThreshold', new_derivative_threshold)


def get_feature_names() -> list[str]:
    """Return a list with the name of all the available features

    Returns:
        A list that contains all the feature names available in
        the eFEL. These names can be used in the feature_names
        argument of e.g. get_feature_values()
    """
    cppcore.Initialize(_settings.dependencyfile_path, "log")
    feature_names: list[str] = []
    cppcore.getFeatureNames(feature_names)

    feature_names += pyfeatures.all_pyfeatures

    return feature_names


def feature_name_exists(feature_name: str) -> bool:
    """Returns True if the feature name exists in eFEL, False otherwise."""
    return feature_name in get_feature_names()


def _get_feature(feature_name: str, raise_warnings=False) -> np.ndarray | None:
    """Get feature value, decide to use python or cpp"""
    if feature_name in pyfeatures.all_pyfeatures:
        return get_py_feature(feature_name)
    else:
        return get_cpp_feature(feature_name, raise_warnings=raise_warnings)


def get_distance(
        trace: dict,
        feature_name: str,
        mean: float,
        std: float,
        trace_check: bool = True,
        error_dist: float = 250) -> float:
    """Calculate distance value for a list of traces.

    Args:
        trace: Trace dict that represents one trace. The dict should have the
               following keys: 'T', 'V', 'stim_start', 'stim_end'
        feature_name: Name of the the features for which to calculate the distance
        mean: Mean to calculate the distance from
        std: Std to scale the distance with
        trace_check: Let the library check if there are spikes outside of stimulus
                     interval, default is True
        error_dist: Distance returned when error, default is 250

    Returns:
        The absolute number of standard deviation the feature is away
        from the mean. In case of anomalous results a value of
        'error_dist' standard deviations is returned.
        This can happen if: a feature generates an error, there are
        spikes outside of the stimulus interval, the feature returns
        a NaN, etc.
    """
    _initialise()

    # Next set time, voltage and the stimulus start and end
    for item in list(trace.keys()):
        cppcore.setFeatureDouble(item, [x for x in trace[item]])

    if trace_check:
        trace_check_success = get_feature_values(
            [trace], ['trace_check'], None, True, True)[0]

        if trace_check_success["trace_check"] is None:
            return error_dist

    feature_values = _get_feature(feature_name)

    distance = 0.0
    if feature_values is None or len(feature_values) < 1:
        return error_dist
    else:
        # Am not using anything more fancy to avoid breaking exact
        # reproducibility of legacy C++ code
        for feature_value in feature_values:
            distance += abs(feature_value - mean)

        distance = distance / std / len(feature_values)

        # Check for NaN
        if distance != distance:
            return error_dist

        return distance


def _initialise() -> None:
    """Set cppcore initial values."""
    cppcore.Initialize(_settings.dependencyfile_path, "log")
    # flush the GErrorString from previous runs by calling getgError()
    cppcore.getgError()
    _initSettings()


def _initSettings() -> None:
    """Init the settings in cppcore."""
    settings_attrs = vars(_settings)
    for setting_name, setting_value in settings_attrs.items():
        if isinstance(setting_value, bool):
            setting_value = int(setting_value)
        if isinstance(setting_value, int):
            cppcore.setFeatureInt(setting_name, [setting_value])
        elif isinstance(setting_value, float):
            if isinstance(setting_value, list):
                cppcore.setFeatureDouble(setting_name, setting_value)
            else:
                cppcore.setFeatureDouble(setting_name, [setting_value])
        elif isinstance(setting_value, str):
            cppcore.setFeatureString(setting_name, setting_value)


@deprecated("Use `set_setting()` instead")
def set_int_setting(setting_name: str, new_value: int) -> None:
    """Set a certain integer setting to a new value."""
    set_setting(setting_name, new_value)


@deprecated("Use `set_setting()` instead")
def set_double_setting(setting_name: str, new_value: float) -> None:
    """Set a certain double setting to a new value."""
    set_setting(setting_name, new_value)


@deprecated("Use `set_setting()` instead")
def set_str_setting(setting_name: str, new_value: str) -> None:
    """Set a certain string setting to a new value."""
    set_setting(setting_name, new_value)


@overload
def get_feature_values(
    traces: list[dict],
    feature_names: list[str],
    parallel_map: Callable | None,
    return_list: Literal[True],
    raise_warnings: bool = True,
) -> list:
    ...


@overload
def get_feature_values(
    traces: list[dict],
    feature_names: list[str],
    parallel_map: Callable | None,
    return_list: Literal[False],
    raise_warnings: bool = True,
) -> Iterator:
    ...


def get_feature_values(
    traces: list[dict],
    feature_names: list[str],
    parallel_map: Callable | None = None,
    return_list: bool = True,
    raise_warnings: bool = True,
) -> list | Iterator:
    """Calculate feature values for a list of traces.

    This function is the core of eFEL API. A list of traces (in the form
    of dictionaries) is passed as argument, together with a list of feature
    names.

    The return value consists of a list of dictionaries, one for each input
    trace. The keys in the dictionaries are the names of the calculated
    features, the corresponding values are lists with the feature values.
    Beware that every feature returns an array of values. E.g. AP_amplitude
    will return a list with the amplitude of every action potential.

    Args:
        traces: Every trace dict represents one trace. The dict should have the
                following keys: 'T', 'V', 'stim_start', 'stim_end'
        feature_names: List with the names of the features to be calculated on all
                       the traces.
        parallel_map: Map function to parallelise over the traces. Default is the
                      serial map() function
        return_list: By default the function returns a list of dicts. This
                     optional argument can disable this, so that the result of the
                     parallel_map() is returned. Can be useful for performance
                     reasons when an iterator is preferred.
        raise_warnings: Raise warning when efel c++ returns an error

    Returns:
        For every input trace a feature value dict is returned (in
        the same order). The dict contains the keys of
        'feature_names', every key contains a numpy array with
        the feature values returned by the C++ efel code.
        The value is None if an error occured during the
        calculation of the feature.
    """

    if parallel_map is None:
        parallel_map = map

    traces_featurenames = ((trace, feature_names, raise_warnings) for trace in traces)
    map_result = parallel_map(_get_feature_values_serial, traces_featurenames)

    if return_list:
        return list(map_result)
    else:
        return map_result


def get_py_feature(feature_name: str) -> np.ndarray | None:
    """Return values of the given feature name."""
    return getattr(pyfeatures, feature_name)()


def _get_feature_values_serial(
    trace_featurenames: tuple[dict, list[str], bool]
) -> dict:
    """Single process of get_feature_values."""
    trace, feature_names, raise_warnings = trace_featurenames
    result = {}

    if 'stim_start' in trace and 'stim_end' in trace:
        try:
            len(trace['stim_start'])
            len(trace['stim_end'])
        except BaseException:
            raise Exception('Unable to determine length of stim_start or '
                            'stim_end, are you sure these are lists ?')

        if len(trace['stim_start']) == 1 and len(trace['stim_end']) == 1:
            if trace['stim_end'][0] <= trace['stim_start'][0]:
                raise Exception(
                    'stim_end needs to be larger than '
                    'stim_start:\nstim_start=%f stim_end=%f' %
                    (trace['stim_start'][0], trace['stim_end'][0]))
        else:
            raise Exception(
                'stim_start and stim_end in the trace '
                'dictionary need to be lists of exactly 1 element')

    else:
        raise Exception('stim_start or stim_end missing from trace')

    cppcore.Clear()
    _initSettings()

    # Next set time, voltage and the stimulus start and end
    for item in list(trace.keys()):
        cppcore.setFeatureDouble(item, [x for x in trace[item]])

    for feature_name in feature_names:
        result[feature_name] = _get_feature(
            feature_name, raise_warnings=raise_warnings)

    return result


def get_mean_feature_values(
        traces: list[dict],
        feature_names: list[str],
        raise_warnings: bool = True) -> list[dict]:
    """Convenience function that returns mean values from get_feature_values()

    Instead of return a list of values for every feature as get_feature_values()
    does, this function returns per trace one value for every feature, namely
    the mean value.

    Args:
        traces: Every trace dict represents one trace. The dict should have the
                following keys: 'T', 'V', 'stim_start', 'stim_end'
        feature_names: List with the names of the features to be calculated on all
                       the traces.
        raise_warnings: Raise warning when efel c++ returns an error

    Returns:
        For every input trace a feature value dict is returned (in
        the same order). The dict contains the keys of
        'feature_names', every key contains the mean of the array
        that is returned by get_feature_values()
        The value is None if an error occured during the
        calculation of the feature, or if the feature value array
        was empty.
    """
    featureDicts = get_feature_values(
        traces,
        feature_names,
        parallel_map=None,
        return_list=True,
        raise_warnings=raise_warnings)
    for featureDict in featureDicts:
        for (key, values) in list(featureDict.items()):
            if values is None or len(values) == 0:
                featureDict[key] = None
            else:
                featureDict[key] = np.mean(values)

    return featureDicts


def register_feature(feature_function: Callable):
    """Register a new feature."""
    pyfeatures.all_pyfeatures.append(feature_function.__name__)
    setattr(pyfeatures, feature_function.__name__, feature_function)


reset()


# Deprecated functions
@deprecated("Use set_threshold instead")
def setThreshold(newThreshold: float) -> None:
    set_threshold(newThreshold)


@deprecated("Use set_derivative_threshold instead")
def setDerivativeThreshold(newDerivativeThreshold: float) -> None:
    set_derivative_threshold(newDerivativeThreshold)


@deprecated("Use get_feature_names instead")
def getFeatureNames() -> list[str]:
    return get_feature_names()


@deprecated("Use feature_name_exists instead")
def FeatureNameExists(feature_name: str) -> bool:
    return feature_name_exists(feature_name)


@deprecated("Use get_distance instead")
def getDistance(
        trace,
        featureName,
        mean,
        std,
        trace_check=True,
        error_dist=250) -> float:
    return get_distance(trace, featureName, mean, std, trace_check, error_dist)


@deprecated("Use set_int_setting instead")
def setIntSetting(setting_name: str, new_value: int) -> None:
    set_int_setting(setting_name, new_value)


@deprecated("Use set_double_setting instead")
def setDoubleSetting(setting_name: str, new_value: float) -> None:
    set_double_setting(setting_name, new_value)


@deprecated("Use set_str_setting instead")
def setStrSetting(setting_name: str, new_value: str) -> None:
    set_str_setting(setting_name, new_value)


@deprecated("Use get_feature_values instead")
def getFeatureValues(
        traces,
        featureNames,
        parallel_map=None,
        return_list=True,
        raise_warnings=True):
    return get_feature_values(
        traces, featureNames, parallel_map, return_list, raise_warnings)


@deprecated("Use get_mean_feature_values instead")
def getMeanFeatureValues(
        traces,
        featureNames,
        raise_warnings=True):
    return get_mean_feature_values(traces, featureNames, raise_warnings)


@deprecated("Use get_dependency_file_location instead")
def getDependencyFileLocation() -> str:
    return get_dependency_file_location()


@deprecated("Use set_dependency_file_location instead")
def setDependencyFileLocation(location: str | Path) -> None:
    return set_dependency_file_location(location)



"""Module containing access functions to C++ features for Python features."""
from __future__ import annotations
import warnings
import numpy as np
from efel import cppcore


def get_cpp_feature(feature_name: str, raise_warnings=False) -> np.ndarray | None:
    """Return value of feature implemented in cpp."""
    cppcoreFeatureValues: list[int | float] = list()
    exitCode = cppcore.getFeature(feature_name, cppcoreFeatureValues)
    if exitCode < 0:
        if raise_warnings:
            warnings.warn(
                f"Error while calculating {feature_name}, {cppcore.getgError()}",
                RuntimeWarning)
        return None
    return np.array(cppcoreFeatureValues)


def _get_cpp_data(data_name: str) -> float | int:
    """Get cpp data value."""
    try:
        return cppcore.getMapDoubleData(data_name)[0]
    except Exception:
        return cppcore.getMapIntData(data_name)[0]



def Initialize(depfilename: str, outfilename: str) -> int: ...
def Clear() -> int: ...
def getFeature(feature_name: str, values: list) -> int: ...
def getFeatureInt(feature_name: str, values: list[int]) -> int: ...
def getFeatureDouble(feature_name: str, values: list[float]) -> int: ...
def getMapIntData(data_name: str) -> list[int]: ...
def getMapDoubleData(data_name: str) -> list[float]: ...
def setFeatureInt(feature_name: str, values: list[int]) -> int: ...
def setFeatureDouble(feature_name: str, values: list[float]) -> float: ...
def setFeatureString(feature_name: str, value: str) -> int: ...
def featuretype(feature_name: str) -> str: ...
def getgError() -> str: ...
def getFeatureNames(feature_names: list[str]) -> None: ...



from __future__ import annotations
"""Python implementation of features"""

"""
Copyright (c) 2015, Blue Brain Project/EPFL
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * Neither the name of the copyright holder nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from efel.pyfeatures.cppfeature_access import _get_cpp_data, get_cpp_feature
from efel.pyfeatures.isi import *
from typing_extensions import deprecated

import numpy as np
from numpy.fft import *

all_pyfeatures = [
    'voltage',
    'time',
    'current',
    'ISIs',
    'ISI_values',
    'ISI_CV',
    'single_burst_ratio',
    'irregularity_index',
    'burst_ISI_indices',
    'burst_mean_freq',
    'interburst_voltage',
    'ISI_log_slope',
    'ISI_semilog_slope',
    'ISI_log_slope_skip',
    'initburst_sahp',
    'initburst_sahp_vb',
    'initburst_sahp_ssse',
    'depol_block',
    'depol_block_bool',
    'Spikecount',
    'Spikecount_stimint',
    'spike_count',
    'spike_count_stimint',
    'spikes_per_burst',
    'spikes_per_burst_diff',
    'spikes_in_burst1_burst2_diff',
    'spikes_in_burst1_burstlast_diff',
    'impedance',
    'burst_number',
    'strict_burst_number',
    'trace_check',
    'phaseslope_max',
    'inv_ISI_values',
    'inv_first_ISI',
    'inv_second_ISI',
    'inv_third_ISI',
    'inv_fourth_ISI',
    'inv_fifth_ISI',
    'inv_last_ISI',
    'activation_time_constant',
    'deactivation_time_constant',
    'inactivation_time_constant',
]


def voltage() -> np.ndarray | None:
    """Get voltage trace."""
    return get_cpp_feature("voltage")


def time() -> np.ndarray | None:
    """Get time trace."""
    return get_cpp_feature("time")


@deprecated("Use spike_count instead.")
def Spikecount() -> np.ndarray:
    return spike_count()


def spike_count() -> np.ndarray:
    """Get spike count."""
    peak_indices = get_cpp_feature("peak_indices")
    if peak_indices is None:
        return np.array([0])
    return np.array([peak_indices.size])


@deprecated("Use spike_count_stimint instead.")
def Spikecount_stimint() -> np.ndarray:
    return spike_count_stimint()


def spike_count_stimint() -> np.ndarray:
    """Get spike count within stimulus interval."""
    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")
    peak_times = get_cpp_feature("peak_time")
    if peak_times is None:
        return np.array([0])

    res = sum(1 for time in peak_times if stim_start <= time <= stim_end)
    return np.array([res])


def trace_check() -> np.ndarray | None:
    """Returns np.array([0]) if there are no spikes outside stimulus boundaries.

    Returns None upon failure.
    """
    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")
    peak_times = get_cpp_feature("peak_time")
    if peak_times is None:  # If no spikes, then no problem
        return np.array([0])
    # Check if there are no spikes or if all spikes are within the stimulus interval
    if np.all((peak_times >= stim_start) & (peak_times <= stim_end * 1.05)):
        return np.array([0])  # 0 if trace is valid
    else:
        return None  # None if trace is invalid due to spike outside stimulus interval


def burst_number() -> np.ndarray:
    """The number of bursts."""
    mean_freq = burst_mean_freq()
    return np.array([0]) if mean_freq is None else np.array([mean_freq.size])


def impedance():
    from scipy.ndimage.filters import gaussian_filter1d

    dt = _get_cpp_data("interp_step")
    Z_max_freq = _get_cpp_data("impedance_max_freq")
    voltage_trace = get_cpp_feature("voltage")
    holding_voltage = get_cpp_feature("voltage_base")
    # when stimulus starts at t=0, use steady_state_voltage_stimend as proxy
    if holding_voltage is None:
        holding_voltage = get_cpp_feature("steady_state_voltage_stimend")
    normalized_voltage = voltage_trace - holding_voltage[0]
    current_trace = current()
    if current_trace is not None:
        holding_current = get_cpp_feature("current_base")
        # when stimulus starts at t=0, use steady_state_current_stimend as proxy
        if holding_current is None:
            holding_current = get_cpp_feature("steady_state_current_stimend")
        normalized_current = current_trace - holding_current[0]
        n_spikes = spike_count()
        if n_spikes < 1:  # if there is no spikes in ZAP
            fft_volt = np.fft.fft(normalized_voltage)
            fft_cur = np.fft.fft(normalized_current)
            if any(fft_cur) == 0:
                return None
            # convert dt from ms to s to have freq in Hz
            freq = np.fft.fftfreq(len(normalized_voltage), d=dt / 1000.)
            Z = fft_volt / fft_cur
            norm_Z = abs(Z) / max(abs(Z))
            select_idxs = np.swapaxes(
                np.argwhere((freq > 0) & (freq <= Z_max_freq)), 0, 1
            )[0]
            smooth_Z = gaussian_filter1d(norm_Z[select_idxs], 10)
            ind_max = np.argmax(smooth_Z)
            return np.array([freq[select_idxs][ind_max]])
        else:
            return None
    else:
        return None


def current():
    """Get current trace"""
    return get_cpp_feature("current")


def initburst_sahp_vb():
    """SlowAHP voltage from voltage base after initial burst"""

    # Required cpp features
    initburst_sahp_value = initburst_sahp()
    voltage_base = get_cpp_feature("voltage_base")

    if initburst_sahp_value is None or voltage_base is None or \
            len(initburst_sahp_value) != 1 or len(voltage_base) != 1:
        return None
    else:
        return np.array([initburst_sahp_value[0] - voltage_base[0]])


def initburst_sahp_ssse():
    """SlowAHP voltage from steady_state_voltage_stimend after initial burst"""

    # Required cpp features
    initburst_sahp_value = initburst_sahp()
    ssse = get_cpp_feature("steady_state_voltage_stimend")

    if initburst_sahp_value is None or ssse is None or \
            len(initburst_sahp_value) != 1 or len(ssse) != 1:
        return None
    else:
        return np.array([initburst_sahp_value[0] - ssse[0]])


def depol_block():
    """Check for a depolarization block"""

    # if there is no depolarization block return 1
    # if there is a depolarization block return None
    # subthreshold traces will also return 1

    # Required trace data
    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")

    # Required cpp features
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")
    AP_begin_voltage = get_cpp_feature("AP_begin_voltage")
    stim_start_idx = np.flatnonzero(time >= stim_start)[0]
    stim_end_idx = np.flatnonzero(time >= stim_end)[0]

    if AP_begin_voltage is None:
        return np.array([1])  # if subthreshold no depolarization block
    elif AP_begin_voltage.size:
        depol_block_threshold = np.mean(AP_begin_voltage)  # mV
    else:
        depol_block_threshold = -50

    block_min_duration = 50.0  # ms
    long_hyperpol_threshold = -75.0  # mV

    bool_voltage = np.array(voltage > depol_block_threshold, dtype=int)
    up_indexes = np.flatnonzero(np.diff(bool_voltage) == 1)
    down_indexes = np.flatnonzero(np.diff(bool_voltage) == -1)
    if len(up_indexes) > len(down_indexes):
        down_indexes = np.append(down_indexes, [stim_end_idx])

    if len(up_indexes) == 0:
        # if it never gets high enough, that's not a good sign (meaning no
        # spikes)
        return None
    else:
        # if it stays in the depolarization block more than min_duration, flag
        # as depolarization block
        max_depol_duration = np.max(
            [time[down_indexes[k]] - time[up_idx] for k,
             up_idx in enumerate(up_indexes)])
        if max_depol_duration > block_min_duration:
            return None

    bool_voltage = np.array(voltage > long_hyperpol_threshold, dtype=int)
    up_indexes = np.flatnonzero(np.diff(bool_voltage) == 1)
    down_indexes = np.flatnonzero(np.diff(bool_voltage) == -1)
    down_indexes = down_indexes[(down_indexes > stim_start_idx) & (
        down_indexes < stim_end_idx)]
    if len(down_indexes) != 0:
        up_indexes = up_indexes[(up_indexes > stim_start_idx) & (
            up_indexes < stim_end_idx) & (up_indexes > down_indexes[0])]
        if len(up_indexes) < len(down_indexes):
            up_indexes = np.append(up_indexes, [stim_end_idx])
        max_hyperpol_duration = np.max(
            [time[up_indexes[k]] - time[down_idx] for k,
             down_idx in enumerate(down_indexes)])

        # if it stays in hyperpolarized stage for more than min_duration,
        # flag as depolarization block
        if max_hyperpol_duration > block_min_duration:
            return None

    return np.array([1])


def depol_block_bool():
    """Wrapper around the depol_block feature. Returns [1] if depol_block
    is None, [0] otherwise."""

    if depol_block() is None:
        return np.array([1])
    else:
        return np.array([0])


def spikes_per_burst():
    """Calculate the number of spikes per burst"""

    burst_begin_indices = get_cpp_feature("burst_begin_indices")
    burst_end_indices = get_cpp_feature("burst_end_indices")

    if burst_begin_indices is None or len(burst_begin_indices) < 1:
        return None

    ap_per_bursts = []
    for idx_begin, idx_end in zip(burst_begin_indices, burst_end_indices):
        ap_per_bursts.append(idx_end - idx_begin + 1)

    return np.array(ap_per_bursts)


def spikes_per_burst_diff():
    """Calculate the diff between the spikes in each burst and the next one"""
    spikes_per_burst_values = spikes_per_burst()
    if spikes_per_burst_values is None or len(spikes_per_burst_values) < 2:
        return None

    return spikes_per_burst_values[:-1] - spikes_per_burst_values[1:]


def spikes_in_burst1_burst2_diff():
    """Calculate the diff between the spikes in 1st and 2nd bursts"""
    spikes_per_burst_diff_values = spikes_per_burst_diff()
    if spikes_per_burst_diff_values is None or len(
        spikes_per_burst_diff_values
    ) < 1:
        return None

    return np.array([spikes_per_burst_diff_values[0]])


def spikes_in_burst1_burstlast_diff():
    """Calculate the diff between the spikes in 1st and last bursts"""
    spikes_per_burst_values = spikes_per_burst()
    if spikes_per_burst_values is None or len(spikes_per_burst_values) < 2:
        return None

    return np.array([
        spikes_per_burst_values[0] - spikes_per_burst_values[-1]
    ])


def phaseslope_max() -> np.ndarray | None:
    """Calculate the maximum phase slope."""
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")
    if voltage is None or time is None:
        return None
    time = time[:len(voltage)]

    from numpy import diff

    phaseslope = diff(voltage) / diff(time)
    try:
        return np.array([np.max(phaseslope)])
    except ValueError:
        return None


def exp_fit(t, tau, A0, A1) -> np.ndarray | float:
    """Exponential function used in exponential fitting.

    Args:
        t (ndarray or float): time series
        tau (float): time constant
        A0 (float): constant added to the exponential
        A1 (float): constant multiplying the exponential
    """
    return A0 + A1 * np.exp(-t / tau)


def activation_time_constant() -> np.ndarray | None:
    """Time constant for an ion channel activation trace.
    Fits for stim_start to trace maximum interval as A - B * exp(-t/tau)."""
    from scipy.optimize import curve_fit

    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")

    if voltage is None or time is None:
        return None

    # isolate stimulus interval
    stim_start_idx = np.flatnonzero(time >= stim_start)[0]
    stim_end_idx = np.flatnonzero(time >= stim_end)[0]
    time_interval = time[stim_start_idx:stim_end_idx]
    voltage_interval = voltage[stim_start_idx:stim_end_idx]

    # keep trace going from stim_start to voltage max
    max_idx = np.argmax(voltage_interval)
    time_interval = time_interval[:max_idx + 1]
    voltage_interval = voltage_interval[:max_idx + 1]

    # correct time so that it starts from 0
    time_interval -= time_interval[0]

    # fit
    try:
        popt, _ = curve_fit(
            exp_fit,
            time_interval,
            voltage_interval,
            p0=(1., voltage_interval[-1], voltage_interval[0] - voltage_interval[-1]),
            # positive tau, negative A1
            bounds=((0, -np.inf, -np.inf), (np.inf, np.inf, 0)),
            nan_policy="omit",
        )
    except (ValueError, RuntimeError):
        return None

    return np.array([abs(popt[0])])


def deactivation_time_constant() -> np.ndarray | None:
    """Time constant for an ion channel deactivation trace.
    Fits for stim_start to stim_end as A + B * exp(-t/tau)."""
    from scipy.optimize import curve_fit

    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")

    if voltage is None or time is None:
        return None

    # isolate stimulus interval
    interval_indices = np.where((time >= stim_start) & (time < stim_end))
    time_interval = time[interval_indices]
    voltage_interval = voltage[interval_indices]

    # correct time so that it starts from 0
    time_interval -= time_interval[0]

    # fit
    try:
        popt, _ = curve_fit(
            exp_fit,
            time_interval,
            voltage_interval,
            p0=(
                1., voltage_interval[-1], max(
                    0, voltage_interval[0] - voltage_interval[-1]
                )
            ),
            bounds=((0, -np.inf, 0), np.inf),  # positive tau, positive A1
            nan_policy="omit",
        )
    except (ValueError, RuntimeError):
        return None

    return np.array([abs(popt[0])])


def inactivation_time_constant() -> np.ndarray | None:
    """Time constant for an ion channel inactivation trace.
    Fits for trace maximum to stim end interval as A + B * exp(-t/tau)."""
    from scipy.optimize import curve_fit

    stim_start = _get_cpp_data("stim_start")
    stim_end = _get_cpp_data("stim_end")
    voltage = get_cpp_feature("voltage")
    time = get_cpp_feature("time")
    # used to remove end of trace to remove artifacts due to stimulus change
    end_skip = _get_cpp_data("inactivation_tc_end_skip")

    if voltage is None or time is None:
        return None

    # isolate stimulus interval
    stim_start_idx = np.flatnonzero(time >= stim_start)[0]
    stim_end_idx = np.flatnonzero(time >= stim_end)[0]
    time_interval = time[stim_start_idx:stim_end_idx - end_skip]
    voltage_interval = voltage[stim_start_idx:stim_end_idx - end_skip]

    # keep trace going from voltage max to stim end
    # remove end of trace to remove artifacts due to stimulus change
    max_idx = np.argmax(voltage_interval)
    time_interval = time_interval[max_idx:]
    voltage_interval = voltage_interval[max_idx:]

    # correct time so that it starts from 0
    if time_interval.size < 1:
        return None
    time_interval -= time_interval[0]

    # fit
    try:
        popt, _ = curve_fit(
            exp_fit,
            time_interval,
            voltage_interval,
            p0=(1., voltage_interval[-1], voltage_interval[0] - voltage_interval[-1]),
            bounds=((0, -np.inf, 0), np.inf),  # positive tau, positive A1
            nan_policy="omit",
        )
    except (ValueError, RuntimeError):
        return None

    return np.array([abs(popt[0])])



"""efel Settings class"""

"""
Copyright (c) 2015, EPFL/Blue Brain Project

 This file is part of eFEL <https://github.com/BlueBrain/eFEL>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from dataclasses import dataclass, fields, asdict
from pathlib import Path
from typing import Union
from typing_extensions import deprecated
import logging

logger = logging.getLogger(__name__)


@dataclass
class Settings:
    """eFEL settings class.

    Attributes:
        Threshold (float): Spike detection threshold (default: -20.0).
        DerivativeThreshold (float): Threshold value for derivative calculations
        (default: 10.0).
        DownDerivativeThreshold (float): Threshold value for downward derivative
        calculations (default: -12.0).
        dependencyfile_path (str): Path to the dependency file
        (default: 'DependencyV5.txt').
        spike_skipf (float): Fraction of spikes to skip (default: 0.1).
        max_spike_skip (int): Maximum number of spikes to skip (default: 2).
        interp_step (float): Interpolation step (default: 0.1).
        burst_factor (float): Burst factor (default: 1.5).
        strict_burst_factor (float): Strict burst factor (default: 2.0).
        voltage_base_start_perc (float): Voltage base start percentage (default: 0.9).
        voltage_base_end_perc (float): Voltage base end percentage (default: 1.0).
        current_base_start_perc (float): Current base start percentage (default: 0.9).
        current_base_end_perc (float): Current base end percentage (default: 1.0).
        rise_start_perc (float): Rise start percentage (default: 0.0).
        rise_end_perc (float): Rise end percentage (default: 1.0).
        initial_perc (float): Initial percentage (default: 0.1).
        min_spike_height (float): Minimum spike height (default: 20.0).
        strict_stiminterval (bool): Strict stimulus interval (default: False).
        initburst_freq_threshold (int): Initial burst frequency threshold
        (default: 50)
        initburst_sahp_start (int): Initial burst SAHP start (default: 5).
        initburst_sahp_end (int): Initial burst SAHP end (default: 100).
        DerivativeWindow (int): Derivative window (default: 3).
        voltage_base_mode (str): Voltage base mode (default: "mean").
        current_base_mode (str): Current base mode (default: "mean").
        precision_threshold (float): Precision threshold (default: 1e-10).
        sahp_start (float): SAHP start (default: 5.0).
        ignore_first_ISI (bool): Ignore first ISI (default: True).
        impedance_max_freq (float): Impedance maximum frequency (default: 50.0).
        inactivation_tc_end_skip (int): number of data points to skip before
            stim end for inactivation_time_constant feature
    """

    Threshold: float = -20.0
    DerivativeThreshold: float = 10.0
    DownDerivativeThreshold: float = -12.0
    dependencyfile_path: str = str(
        Path(__file__).parent.absolute() / "DependencyV5.txt"
    )

    spike_skipf: float = 0.1
    max_spike_skip: int = 2
    interp_step: float = 0.1
    burst_factor: float = 1.5
    strict_burst_factor: float = 2.0
    voltage_base_start_perc: float = 0.9
    voltage_base_end_perc: float = 1.0
    current_base_start_perc: float = 0.9
    current_base_end_perc: float = 1.0
    rise_start_perc: float = 0.0
    rise_end_perc: float = 1.0
    initial_perc: float = 0.1
    min_spike_height: float = 20.0
    strict_stiminterval: bool = False
    initburst_freq_threshold: int = 50
    initburst_sahp_start: int = 5
    initburst_sahp_end: int = 100
    DerivativeWindow: int = 3
    voltage_base_mode: str = "mean"
    current_base_mode: str = "mean"
    precision_threshold: float = 1e-10
    sahp_start: float = 5.0
    ignore_first_ISI: bool = True
    impedance_max_freq: float = 50.0
    AP_phaseslope_range: int = 2
    inactivation_tc_end_skip: int = 10

    def set_setting(self,
                    setting_name: str,
                    new_value: Union[int, float, str, bool]) -> None:
        """Set a certain setting to a new value.

        Args:
            setting_name (str): Name of the setting to be modified.
            new_value (Union[int, float, str, bool]): New value for the setting.

        Raises:
            ValueError: If the value is of the wrong type.
            FileNotFoundError: If the path to the dependency file does not exist
            (for 'dependencyfile_path' setting).
        """
        if hasattr(self, setting_name):
            expected_types = {f.name: f.type for f in fields(self)}
            expected_type = expected_types.get(setting_name)
            if expected_type is None:
                raise TypeError(f"type not found for setting '{setting_name}'")
            try:
                converted_value = expected_type(new_value)
                if not isinstance(new_value, expected_type):
                    log_message = (
                        "Value '%s' of type '%s' for setting '%s' "
                        "has been converted to '%s' of type '%s'."
                    ) % (
                        new_value,
                        type(new_value).__name__,
                        setting_name,
                        converted_value,
                        expected_type.__name__
                    )
                    if expected_type is int and isinstance(new_value, float) and \
                       new_value != converted_value:
                        logger.warning(log_message)
                    else:
                        logger.debug(log_message)
            except (ValueError, TypeError):
                raise ValueError(f"Invalid value for setting '{setting_name}'. "
                                 f"Expected type: {expected_type.__name__}.")
        else:
            logger.debug("Setting '%s' not found in settings. "
                         "Adding it as a new setting.", setting_name)
            converted_value = new_value

        if setting_name == "dependencyfile_path":
            path = Path(str(converted_value))
            if not path.exists():
                raise FileNotFoundError(f"Path to dependency file {converted_value}"
                                        "doesn't exist")

        setattr(self, setting_name, converted_value)

    def reset_to_default(self):
        """Reset settings to their default values"""
        default_settings = Settings()
        for field in default_settings.__dataclass_fields__:
            setattr(self, field, getattr(default_settings, field))

    def __str__(self):
        attributes = asdict(self)
        return '\n'.join([f"{key}: {value}" for key, value in attributes.items()])
